import json
import importlib

from typing import Optional
from ovos_utils.log import LOG, log_deprecation
from os import makedirs
from os.path import join, dirname
from ovos_utils.xdg_utils import xdg_config_home

try:
    # TODO: deprecate in 0.0.12, leave only during 0.0.11 life cycle
    from ovos_utils.file_utils import FileWatcher as _FileWatcher, \
        FileEventHandler as _FileEventHandler


    class FileWatcher(_FileWatcher):
        def __init__(self, *args, **kwargs):
            if _FileWatcher is None:
                raise ImportError("Import from ovos_utils.file_utils directly")
            _FileWatcher.__init__(self, *args, **kwargs)
            log_deprecation("Import from ovos_utils.file_utils directly",
                            "0.0.12")


    class FileEventHandler(_FileEventHandler):
        def __init__(self, *args, **kwargs):
            if _FileEventHandler is None:
                raise ImportError("Import from ovos_utils.file_utils directly")
            _FileEventHandler.__init__(self, *args, **kwargs)
            log_deprecation("Import from ovos_utils.file_utils directly",
                            "0.0.12")

except ImportError:
    LOG.debug("Failed to import `FileWatcher` and `FileEventHandler`")
    FileWatcher = None
    FileEventHandler = None


def init_module_config(module_name: str, module_override: str,
                       module_config: Optional[dict] = None):
    """
    Configure a module to use a configured override
    @param module_name: module to configure
    @param module_override: key in `module_overrides` to assign to `module_name`
    @param module_config: configuration for the specified module_override
    """
    from ovos_config.meta import get_ovos_config
    ovos_conf = get_ovos_config()
    ovos_conf.setdefault('module_overrides', {})
    ovos_conf.setdefault('submodule_mappings', {})

    config_updated = False

    # Check for and update module_overrides
    existing_config = ovos_conf['module_overrides'].get(module_override)
    if module_config:
        if existing_config and module_config != existing_config:
            LOG.info(f"existing={existing_config}")
            LOG.warning(f"requested={module_override}")
            raise RuntimeError("Passed configuration conflicts with existing")
        elif not existing_config:
            LOG.info(f"Configuring {module_override} config")
            ovos_conf['module_overrides'][module_override] = module_config
            config_updated = True

    # Check for and update submodules
    if module_name == "__main__":
        raise ValueError("Configuring `__main__` has unintended consequences"
                         "and is not supported here")
    if module_name in ovos_conf['submodule_mappings']:
        LOG.debug(f"{module_name} already configured, skipping configuration")
    else:
        ovos_conf["submodule_mappings"][module_name] = module_override
        config_updated = True

    # Check if ovos config is modified
    if not config_updated:
        LOG.debug("Configuration unchanged")
        return

    ovos_path = join(xdg_config_home(), "OpenVoiceOS", "ovos.conf")
    makedirs(dirname(ovos_path), exist_ok=True)
    config_to_write = {
        "module_overrides": ovos_conf.get('module_overrides'),
        "submodule_mappings": ovos_conf.get('submodule_mappings')
    }
    LOG.info(f"Updating configuration at: {ovos_path}")
    with open(ovos_path, "w+") as f:
        json.dump(config_to_write, f, indent=4)

    # Note that the below block reloads modules in a specific order due to
    # imports within ovos_config and mycroft.configuration
    import ovos_config
    importlib.reload(ovos_config.locations)  # Reset global config paths
    # from ovos_config.meta import get_ovos_config
    ovos_conf = get_ovos_config()  # Load the full stack for /etc overrides

    module_default_config = ovos_conf["module_overrides"][module_override].get(
        "default_config_path")
    if module_default_config \
            and ovos_config.locations.DEFAULT_CONFIG != module_default_config:
        ovos_config.locations.DEFAULT_CONFIG = module_default_config

        # Default config changed, remove any cached configuration
        del ovos_config.config.Configuration
        del ovos_config.Configuration

    import ovos_config.models
    importlib.reload(ovos_config.models)
    importlib.reload(ovos_config.config)
    importlib.reload(ovos_config)



from dateutil.tz import gettz, tzlocal

import ovos_config

# lingua_franca is optional and might not be installed
# exceptions should only be raised in the parse and format utils


try:
    import lingua_franca as LF
except ImportError:
    LF = None

_lang = None
_default_tz = None


def get_full_lang_code(lang):
    """ given a 2-letter lang code, return the full default 4-letter code"""
    # first give preference to any configured dialects
    # eg, pt-br instead of pt-pt
    valid_langs = get_valid_languages()
    for l in valid_langs:
        if l.split("-")[0] == lang:
            return l

    # just go with the default full code
    langmap = {'az': 'az-az',
               'ca': 'ca-es',
               'cs': 'cs-cz',
               'da': 'da-dk',
               'de': 'de-de',
               'en': 'en-us',
               'es': 'es-es',
               'eu': 'eu-eu',
               'fa': 'fa-ir',
               'fr': 'fr-fr',
               'hu': 'hu-hu',
               'it': 'it-it',
               'nl': 'nl-nl',
               'pl': 'pl-pl',
               'pt': 'pt-pt',
               'ru': 'ru-ru',
               'sl': 'sl-si',
               'sv': 'sv-se',
               'tr': 'tr-tr',
               'uk': 'uk-ua'}
    return langmap.get(lang)


def get_primary_lang_code(config=None):
    global _lang
    if not _lang:
        config = config or ovos_config.Configuration()
        _lang = config.get("lang", "en-us")
    return _lang.split("-")[0]


def get_default_lang(config=None):
    """
    Get the default localized language code (i.e. `en-us`)
    @param config: Configuration (default is Configuration())
    @return: lowercase BCP-47 language code
    """
    global _lang
    if not _lang:
        config = config or ovos_config.Configuration()
        _lang = config.get("lang", "en-us")
    return _lang


def set_default_lang(lang):
    """ setup default language across OVOS packages
    
    currently only configures lingua-franca language, in the future 
    other hooks may be added if we need to perform this operation globally"""
    global _lang
    _lang = lang
    if LF:
        try:
            LF.set_default_lang(lang)
        except:
            pass

def get_config_tz():
    code = ovos_config.Configuration()["location"]["timezone"]["code"]
    return gettz(code)


def get_default_tz():
    # if default was set at runtime use it else use the timezone from .conf
    return _default_tz or get_config_tz()


def set_default_tz(tz=None):
    """ configure timezone across OVOS packages
    
    currently only configures lingua-franca, in the future 
    other hooks may be added if we need to perform this operation globally """
    global _default_tz
    tz = tz or get_config_tz() or tzlocal()
    _default_tz = tz
    if LF:
        # tz added in recently, depends on version
        try:
            LF.time.set_default_tz(tz)
        except:
            pass


def load_languages(langs):
    """ load and configure lang specific resources across OVOS packages
    
    currently only loads lingua-franca language data, in the future 
    other hooks may be added if we need to perform this operation globally"""
    if LF:
        try:
            LF.load_languages(langs)
        except:
            pass


def load_language(lang):
    """ load and configure lang specific resources across OVOS packages
    
    currently only loads lingua-franca language data, in the future 
    other hooks may be added if we need to perform this operation globally"""
    if LF:
        try:
            LF.load_language(lang)
        except:
            pass


def get_valid_languages():
    """ return all valid runtime languages according to mycroft.conf """
    lang_code = ovos_config.Configuration().get("lang", "en-us")
    extra_lang_codes = ovos_config.Configuration().get("secondary_langs", [])
    return set([lang_code] + extra_lang_codes)


def setup_locale(lang=None, tz=None):
    """ setup default language, timezone and other locale data across OVOS packages
    
    currently only configures lingua-franca, in the future 
    other hooks may be added if we need to perform this operation globally"""
    lang_code = lang or ovos_config.Configuration().get("lang", "en-us")
    valid_langs = get_valid_languages()
    # load any lang specific resources
    load_languages(valid_langs)
    # Set the active lang to match the configured one
    set_default_lang(lang_code)
    # Set the default timezone to match the configured one
    set_default_tz(tz)


# mycroft-core backwards compat LF only interface
def set_default_lf_lang(lang_code="en-us"):
    """Set the default language of Lingua Franca for parsing and formatting.

    Note: this is a temporary method until a global set_default_lang() method
    can be implemented that updates all Mycroft systems eg STT and TTS.
    It will be deprecated at the earliest possible point.

    Args:
        lang (str): BCP-47 language code, e.g. "en-us" or "es-mx"
    """
    return set_default_lang(lang_code)



from ovos_config.config import Configuration, LocalConf, RemoteConf
from ovos_config.locale import set_default_lf_lang, setup_locale, \
    set_default_tz, set_default_lang, get_default_tz, get_default_lang, \
    get_config_tz, get_primary_lang_code, load_languages, load_language
from ovos_config.locations import SYSTEM_CONFIG, USER_CONFIG, DEFAULT_CONFIG, get_xdg_config_locations
from ovos_config.meta import get_ovos_config
